/*
 * Copyright (c) 2023 University of Michigan
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this
 * software and associated documentation files (the “Software”), to deal in the Software
 * without restriction, including without limitation the rights to use, copy, modify,
 * merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or
 * substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
 * INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#ifndef GREEN_AC_EXCEPT_H
#define GREEN_AC_EXCEPT_H

#include <stdexcept>

namespace green::ac {

  class ac_data_error : public std::runtime_error {
  public:
    explicit ac_data_error(const std::string& basic_string) : runtime_error(basic_string) {}
  };

  class ac_data_shape_error : public std::runtime_error {
  public:
    explicit ac_data_shape_error(const std::string& basic_string) : runtime_error(basic_string) {}
  };


  class ac_nevanlinna_error : public std::runtime_error {
  public:
    explicit ac_nevanlinna_error(const std::string& basic_string) : runtime_error(basic_string) {}
  };
}  // namespace green::ac

#endif  // GREEN_AC_EXCEPT_H
